/* =============================================================================
 *       Filename:  Ripple.cpp
 *    Description:  the ripple class stores a variable number of fft windows and
 *                  draws them a wave propagating out from a central point.
 *        Created:  10/30/2011 23:30:28
 *
 *         Author:  Derek Tingle, derek.tingle@gmail.com
 * ===========================================================================*/

#ifdef __MACOSX_CORE__
  #include <GLUT/glut.h>
#else
  #include <GL/gl.h>
  #include <GL/glu.h>
  #include <GL/glut.h>
#endif

#include <math.h>
#include <iostream>
#include <stdio.h>
#include <string.h>
#include <deque>
#include "chuck_fft.h"
#include "Definitions.h"
#include "Ripple.h"

// empty constructor
Ripple::Ripple(){
  m_fft_buff = NULL;
  // deque initialized in header...
  //m_buff_buff = new deque<complex *>;
  //
  // number of positive frequency fft bins
  m_fft_size = 0;
  m_num_windows = 0;
  m_wave_decay = 0;
  m_radius = 0;
  m_dull_colors = NULL;
  m_bright_colors = NULL;
  // stores y scale factor for each delayed fft buffer
  m_y_sin_scale = NULL;
  m_mel_angles = NULL;
  m_use_mel = false;
  m_num_sines=NUM_RIPPLESIN;
}

// real constructor
Ripple::Ripple(complex * fft_buff, unsigned int buff_size, unsigned int num_windows, GLfloat wave_decay, Mutex * mutex){
  m_fft_buff = fft_buff;
  // number of positive frequency fft bins
  m_fft_size = buff_size*ZPF/2;
  m_num_windows = num_windows;
  m_wave_decay = wave_decay;
  m_radius = 1/(float)num_windows;

  m_dull_colors = new RGB_color[m_fft_size];
  m_bright_colors = new RGB_color[m_fft_size];
  Ripple::init_colors();
  // deque initialized in header...
  // m_buff_buff = new deque<complex *>;

  // stores y scale factor for each delayed fft buffer
  m_num_sines=NUM_RIPPLESIN;
  m_y_sin_scale = new float[m_num_windows];
  init_y_sin_scale();

  m_use_mel = false;
  m_mel_angles = new GLfloat[m_fft_size];
  init_mel();

  // not currently being used, but i left it in to find out what's going on
  m_mutex = * mutex;
}

//destructor
Ripple::~Ripple(){}

/* ===  FUNCTION  ==============================================================
 *         Name:  update
 *  Description:  redraws ripple with new fft. called by glutDisplayFunc. 
 * ===========================================================================*/
void Ripple::update_buffs(){
  complex * temp_buff=new complex[m_fft_size];
  for (unsigned int i=0; i<m_fft_size; i++){
    temp_buff[i]=m_fft_buff[i];
  }

  // push current buffer to deque
  m_buff_buff.push_front(temp_buff);

  // make sure we're not storing more than we need to 
  if (m_buff_buff.size() > m_num_windows){
    int to_pop= m_buff_buff.size() - m_num_windows;
    for (int i=0; i < to_pop; i++){
      m_buff_buff.pop_back();
    }
  }
}

/* ===  FUNCTION  ==============================================================
 *         Name:  draw_inner_circle
 *  Description:  takes a buffer of complex fft vals and a raidius and draws a 
 *                a 3d surface where x,z coords are on a circle and y coresponds
 *                to fft magnitude
 * ===========================================================================*/

void Ripple::draw_inner_circle(){
  GLfloat x,z=0;
  float angle=2*M_PI/m_fft_size;
  float cur_angle;
  RGB_color cur_color(0,0,0);

//  m_mutex.lock();

  glBegin(GL_TRIANGLE_FAN);
  glColor4f( 0, 0, 0, 0.5 );
//  glColor4f( 0, 0, 0, 0 );
  glVertex3f(0,0,0); // origin

  
  // loop through m_fft_size+1 values to close the circle

  for(int i = 0; i <= m_fft_size; i++){
    if (m_use_mel){
      cur_angle=m_mel_angles[i%m_fft_size];
    } else{
      cur_angle=i*angle;
    }
    x=m_radius*sin(cur_angle);
    z=m_radius*cos(cur_angle);
    cur_color=m_bright_colors[i%m_fft_size];
    glColor4f( cur_color.r, cur_color.g, cur_color.b, 1 );
    glVertex3f( x, 2 * ZPF * cmp_abs( m_fft_buff[i%m_fft_size]) * m_y_sin_scale[0], z );
  }


  glEnd();

//  m_mutex.unlock();

}

/* ===  FUNCTION  ==============================================================
 *         Name:  draw_ring
 *  Description:  takes a buffer of complex fft vals and a delay index, d, and
 *                draws a 3d surface with from  d*radius to d+1*radius in the
 *                x,z plane and y vals equal to the fft magnitude
 * ===========================================================================*/
void Ripple::draw_ring(unsigned int d){
  // d is the index of which ring to draw

  GLfloat x1,x2,y1_scale,y2_scale,y1_mag,y2_mag,z1,z2=0;
  GLfloat mag_thresh=0.001;
  float angle=2*M_PI/m_fft_size;
  float cur_angle=angle;
  RGB_color color1(0,0,0);
  RGB_color color2(0,0,0);

//  m_mutex.lock();

  glBegin(GL_TRIANGLE_STRIP);

  // loop through m_fft_size+1 values to close the circle
  //
  // Note: this is pretty inefficient. as soon as i increase the number of
  // fft buffers to keep around or the ZPF, i get some lag. i think there are
  // more efficient ways to do this, (like using drawElements or something,)
  // but i'm a little pressed for time
  for(int i=0; i<=m_fft_size; i++){
    if (m_use_mel){
      cur_angle=m_mel_angles[i%m_fft_size];
    } else{
      cur_angle=i*angle;
    }
    x1=d*m_radius*sin(cur_angle);
    x2=(d+1)*m_radius*sin(cur_angle);

    // maybe include switch for other scalings
    // index from d-1 because d loops from 1 to m_fft_size-1
    y1_scale=m_y_sin_scale[d-1] * 2 * ZPF * pow(m_wave_decay,d-1);
    y2_scale=m_y_sin_scale[d] * 2 * ZPF * pow(m_wave_decay,d);

//    y1_mag=cmp_abs( m_buff_buff[d-1][i%m_fft_size]);
    y1_mag=cmp_abs( m_buff_buff[d-1][i%m_fft_size]);
    y2_mag=cmp_abs( m_buff_buff[d][i%m_fft_size]);

    z1=d*m_radius*cos(cur_angle);
    z2=(d+1)*m_radius*cos(cur_angle);


    if (y1_mag > mag_thresh){
      color1=m_bright_colors[i%m_fft_size];
    }
    else{
      color1=m_dull_colors[i%m_fft_size];
    }

    if (y2_mag > mag_thresh){
      color2=m_bright_colors[i%m_fft_size];
    }
    else{
      color2=m_dull_colors[i%m_fft_size];
    }
    glColor4f( color1.r, color1.g, color1.b, 0.75);
    glVertex3f( x1, y1_scale * y1_mag , z1 );
    glColor4f( color2.r, color2.g, color2.b, 0.75);
    glVertex3f( x2, y2_scale * y2_mag , z2 );

    // other potential colorscheme .... weirdness with alpha
//    color1=m_bright_colors[i%m_fft_size];
//    if (y1_mag > mag_thresh){
//      glColor4f( color1.r, color1.g, color1.b, 1);
//    }
//    else{
//      glColor4f( color1.r, color1.g, color1.b, 0.5);
//    }
//    glVertex3f( x1, y1_scale * y1_mag , z1 );
//
//    if (y2_mag > mag_thresh){
//      glColor4f( color2.r, color2.g, color2.b, 1);
//    }
//    else{
//      glColor4f( color2.r, color2.g, color2.b, 0.5);
//    }
//    glVertex3f( x2, y2_scale * y2_mag , z2 );
  }
  glEnd();

//  m_mutex.unlock();
}

// draw ripple
void Ripple::draw(){
  draw_inner_circle();
  int sz=m_buff_buff.size();
  for (int i=1; i<sz; i++){
    draw_ring(i);
  }
}

// init color_wheel
// TODO change to more colors?
void Ripple::init_colors(){
  float hsv_inc=((float) 6)/m_fft_size;
  int j;
  for(size_t i = 0; i < m_fft_size; ++i){
    //tinker with the index to map low pitches to cool colors and high pitches
    //to warm colors
    j=m_fft_size-1-i;
    j=(j-70)%m_fft_size;
    m_dull_colors[j]=hsv_to_rgb(HSV_color(hsv_inc*i,(float) 1,(float) 0.25));
    m_bright_colors[j]=hsv_to_rgb(HSV_color(hsv_inc*i,(float) 1,(float) 1));
//    m_color_wheel[i]=hsv_to_rgb(HSV_color(hsv_inc*i,(float) 1,(float) 1));
  }
}

void Ripple::init_y_sin_scale(){
  for(int i=0; i < m_num_windows; i++){
    m_y_sin_scale[i]=sin( 2*M_PI*m_num_sines*i/m_num_windows);
    // sin( 2*pi*3*i/m_fft_size)?
    // TODO: multiply each y value by a decay and a function
    // TODO: perhaps have a key to increase frequency of sine wave?

//    cerr << "i: " << m_y_sin_scale[i] << endl;
  }
}

void Ripple::clear_buffs(){
  int sz=m_buff_buff.size();
  for (int i=0; i < sz; i++){
    m_buff_buff.pop_back();
  }
}

// intializes an array of log-spaced angle increments (corresponding to mel
// scale)
void Ripple::init_mel(){
  float freq,mel;
  float nyquist= MY_SRATE/2;
  
  GLfloat max_mel=freq_to_mel( nyquist );

  for (int i=0; i<m_fft_size; i++){
    // not finding frequency of center of bin
    freq=(i/(float)m_fft_size)*nyquist;
    mel=freq_to_mel( freq );
    m_mel_angles[i]=(GLfloat)mel/max_mel*360;
//    cerr << i << " "  << m_mel_angles[i] << endl;
  }
}

// converts frequency to mel scale. (log is the natural log)
GLfloat Ripple::freq_to_mel(float freq){
  return (GLfloat)(1127*log(1+(freq/700)));
}

// set flag to use mel scale and clear the old buffer
void Ripple::set_mel_scale(){
  m_use_mel=true;
  clear_buffs();
}

// set flag to use linear scale and clear the old buffer
void Ripple::set_lin_scale(){
  m_use_mel=false;
  clear_buffs();
}

// change decay
void Ripple::change_decay(float decay){
  if (decay==0){
    m_wave_decay=1;
  }
  m_wave_decay+=decay;
  clear_buffs();
}

// change decay
void Ripple::change_num_sines(int num){
  if (num==0){
    m_num_sines=NUM_RIPPLESIN;
  }
  m_num_sines+=num;
  init_y_sin_scale();

  clear_buffs();
}
