/*
 * =============================================================================
 *
 *       Filename:  Ripple.h
 *    Description:  the ripple class stores a variable number of fft windows and
 *                  draws them a wave propagating out from a central point.
 *                  This file also stores the RGB_color and HSV_color structs
 *
 *        Created:  10/30/2011 22:53:05
 *
 *         Author:  Derek Tingle (dt), derek.tingle@gmail.com
 *
 * =============================================================================
 */

#ifndef __RIPPLE_H__
#define __RIPPLE_H__

#include "Thread.h"
#include "MyColors.h"
#include <deque>


/*-----------------------------------------------------------------------------
 *  Ripple Class
 *----------------------------------------------------------------------------*/
class Ripple{
  public:
    Ripple(complex * fft_buff, unsigned int buff_size, unsigned int num_windows, GLfloat wave_decay, Mutex * mutex);
    Ripple();
    ~Ripple();

  public:
    // update
    void update_buffs();
    // draw ripple
    void draw();
    // set flag to use mel scale
    void set_mel_scale();
    // set flag to use linear scale
    void set_lin_scale();
    // sets number of sine waves to use
    void change_num_sines(int num);
    // sets decay (gain)
    void change_decay(float decay);

  private:
    // draw inner circle
    void draw_inner_circle();
    // draw ring
    void draw_ring(unsigned int delay_indx);
    // init color wheels
    void init_colors();
    // init y-scales to be a sine wave
    void init_y_sin_scale();
    // erase all fft buffers
    void clear_buffs();
    // initialize mel_scale array
    void init_mel();
    // converts a frequency to a mel
    GLfloat freq_to_mel(float freq);

  protected:
    // buffer of positive frequency fft values
    complex * m_fft_buff;
    // queue of previous ffts for "waterfall" / wave spreading effect
    deque<complex *> m_buff_buff;
    // number of positive frequency fft bins
    unsigned int m_fft_size;
    // number of delayed fft buffers
    unsigned int m_num_windows;
    // gain for fft wave decay
    GLfloat m_wave_decay;
    // radius of each fft buffer's "ring"
    GLfloat m_radius;
    // array of color values to use for "active" bins
    RGB_color * m_bright_colors;
    // array of color values to use for "boring" bins
    RGB_color * m_dull_colors;
    // mutex
    Mutex m_mutex;
    // array of values to the height of each fft buffer's ring by
    float * m_y_sin_scale;
    // array of angle increments that correspond to mel-scaled freq intervals
    GLfloat * m_mel_angles;
    // flag to use mel scale
    bool m_use_mel;
    // number of sines to use during ripple
    int m_num_sines;
};
#endif
