//  
//  base code written by Jorge Herrera on 2011-10-26.
//  modified by Derek Tingle sometime after that.
// 
//  Third party code used:
//  - chuck_fft.h/cpp based on the FFT implementation in CARL
//  - Thread.h/cpp from STK 


#ifdef __MACOSX_CORE__
  #include <GLUT/glut.h>
#else
  #include <GL/gl.h>
  #include <GL/glu.h>
  #include <GL/glut.h>
#endif


#include <iostream>
#include <math.h>
#include <string.h>

#include "MyAudio.h"
#include "Definitions.h"
#include "chuck_fft.h"
#include "Thread.h"
#include "Ripple.h"


using namespace std;

// ===========
// = Globals =
// ===========
GLsizei g_width = 800; 
GLsizei g_height = 600;

Mutex g_mutex;

float * g_fftBuff;
float * g_window;
bool g_move;
unsigned int g_buffSize = 512;

// global ripple object
Ripple g_ripple;

//-----------------------------------------------------------------------------
// Defines a point in a 3D space (coords x, y and z)
//-----------------------------------------------------------------------------
struct pt3d
{
  pt3d( GLfloat x, GLfloat y, GLfloat z ) : x(x), y(y), z(z) {};
  
  float x;
  float y;
  float z;
};


// =======================
// = Function prototypes =
// =======================
void idleFunc( );
void displayFunc( );
void reshapeFunc( int width, int height );
void keyboardFunc( unsigned char, int, int );
void mouseFunc( int button, int state, int x, int y );
void specialFunc( int key, int x, int y );
void initialize( );
void changeLookAt( pt3d look_from, pt3d look_to, pt3d head_up );
void drawAxis();
void draw_yelline();
void my_view_rotate();


// ==========================
// = Cammera placement vars =
// ==========================
// Camera control global variables
pt3d g_look_from( 0, 0.1, 1);
pt3d g_look_to( 0, 0, 0 );
pt3d g_head_up( 0, 1, 0 );

// amt to change the camera by
GLfloat g_delta=5;

// instead of rotating the camera, we'll just rotate the model
pt3d g_view_rotate(0,0,0);

// full screen flag
bool g_is_full = false;


// ============
// = GL stuff =
// ============
//-----------------------------------------------------------------------------
// Name: initialize( )
// Desc: sets initial OpenGL states
//     also initializes any application data
//-----------------------------------------------------------------------------
void initialize()
{

  // =================
  // = OpenGL & GLUT =
  // =================

  // set the GL clear color - use when the color buffer is cleared
  glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );
    
  // set the shading model to 'smooth'
  glShadeModel( GL_SMOOTH );
  // enable depth
  glEnable( GL_DEPTH_TEST );
  // set the front faces of polygons
  glFrontFace( GL_CCW );
  // set fill mode
  glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );

  // Enable transparency
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      
  // seed random number generator
  srand( time(NULL) );
}


//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( int w, int h )
{
  // save the new window size
  g_width = (GLsizei)w; 
  g_height = (GLsizei)h;
  // map the view port to the client area
  glViewport( 0, 0, (GLsizei)w, (GLsizei)h );
  // set the matrix mode to project
  glMatrixMode( GL_PROJECTION );
  // load the identity matrix
  glLoadIdentity( );
  // create the viewing frustum
  gluPerspective( 45.0, (GLfloat) w / (GLfloat) h, 0.1, 300.0 );
  // set the matrix mode to modelview
  glMatrixMode( GL_MODELVIEW );
  // load the identity matrix
  glLoadIdentity( );
  // position the view point
  changeLookAt( g_look_from,g_look_to, g_head_up );
}


//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
  switch( key )
  {
  case 'Q':
  case 'q':
    cerr << "Bye!" << endl;
    exit(1);
    break;
  case 'L':
  case 'l':
    g_ripple.set_lin_scale();
    cerr << "Using linear scaling of frequencies" << endl;
    break;
  case 'M':
  case 'm':
    g_ripple.set_mel_scale();
    cerr << "Using mel scaling of frequencies" << endl;
    break;
  case 'f':
  case 'F':
    // fullscreen
    if (g_is_full){
      glutReshapeWindow(800,500);
    }
    else
      glutFullScreen();

    g_is_full=!g_is_full;
    break;
  case 'w':
  case 'W':
    cerr << "increasing the number of ripples" << endl;
    g_ripple.change_num_sines(2);
    break;
  case 'x':
  case 'X':
    cerr << "decreasing the number of ripples" << endl;
    g_ripple.change_num_sines(-2);
    break;
  case 's':
  case 'S':
    cerr << "resetting the number of ripples" << endl;
    g_ripple.change_num_sines(0);
    break;
  case 't':
  case 'T':
    cerr << "increasing the gain of the waves" << endl;
    g_ripple.change_decay(0.005);
    break;
  case 'b':
  case 'B':
    cerr << "decreasing the gain of the waves" << endl;
    g_ripple.change_decay(-0.005);
    break;
  case 'g':
  case 'G':
    cerr << "resetting the gain of the waves to 1" << endl;
    g_ripple.change_decay(0);
    break;
  }
  glutPostRedisplay( );
}



//-----------------------------------------------------------------------------
// Name: mouseFunc( )
// Desc: handles mouse stuff
//-----------------------------------------------------------------------------
void mouseFunc( int button, int state, int x, int y )
{
  if( button == GLUT_LEFT_BUTTON )
  {
    // when left mouse button is down, move left
    if( state == GLUT_DOWN )
    {
    }
    else
    {
    }
  }
  else if ( button == GLUT_RIGHT_BUTTON )
  {
    // when right mouse button down, move right
    if( state == GLUT_DOWN )
    {
    }
    else
    {
    }
  }
  else
  {
  }

  glutPostRedisplay( );
}


//-----------------------------------------------------------------------------
// Name: specialFunc( )
// Desc: handles special function keys
//-----------------------------------------------------------------------------
void specialFunc( int key, int x, int y )
{
  if( key == GLUT_KEY_LEFT)
  {
    g_view_rotate.x += g_delta;
  }
  if( key == GLUT_KEY_RIGHT)
  {
    g_view_rotate.x -= g_delta;
  }
  if( key == GLUT_KEY_DOWN)
  {
    g_look_from.y -= (g_delta/100);
  }
  if( key == GLUT_KEY_UP)
  {
    g_look_from.y += (g_delta/100);
  }
  if( key == GLUT_KEY_PAGE_UP)
  {
    std::cerr << "PageUp arrow";
  }
  if( key == GLUT_KEY_PAGE_DOWN)
  {
    std::cerr << "PageDown arrow";
  }

  glutPostRedisplay( );
}


//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
  // render the scene
  glutPostRedisplay( );
}


//-----------------------------------------------------------------------------
// Name: displayFunc( )
// Desc: callback function invoked to draw the client area
//-----------------------------------------------------------------------------
void displayFunc( )
{
  // clear the color and depth buffers
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  // save the current transformation
  glPushMatrix();
    
    // set the view point (camera)
    changeLookAt( g_look_from, g_look_to, g_head_up );

    glLineWidth( 1.0f );  

    my_view_rotate();
  
    glColor4f(1, 1, 1, 1);
  
    // cast to use the complex data type defined in chuck_fft
    complex * fft = (complex *)g_fftBuff;

    if( g_fftBuff )
    {
      // hey jorge, quick question:
      //  so i tried to pass g_mutex by pointer to my Ripple class and call
      //  the lock and unlock functions from there, but for some reason i
      //  was still getting artifacts in the fft. but when i put the locks
      //  here, it works fine. what gives?
      g_mutex.lock();
      g_ripple.draw();
      g_mutex.unlock();
    }

  // restore state
  glPopMatrix();


  // flush!
  glFlush( );
  // swap the double buffer
  glutSwapBuffers( );
}

//-----------------------------------------------------------------------------
// name: changeLookAt()
// desc: changes the point of view
//-----------------------------------------------------------------------------
void changeLookAt( pt3d look_from, pt3d look_to, pt3d head_up )
{
  gluLookAt(  look_from.x, look_from.y, look_from.z, 
        look_to.x, look_to.y, look_to.z, 
        head_up.x, head_up.y, head_up.z );
}



//-----------------------------------------------------------------------------
// name: drawAxis()
// desc: draw cartesian axis (x,y,z) using (r,g,b) respectively
//-----------------------------------------------------------------------------
void drawAxis()
{
  glPushMatrix();
    
  glBegin( GL_LINES );
  
    // x axis
    glColor4f( 1, 0, 0, 1 );
    glVertex3f( 0, 0, 0 );
    glVertex3f( 1, 0, 0 );

    // y axis
    glColor4f( 0, 1, 0, 1 );
    glVertex3f( 0, 0, 0 );
    glVertex3f( 0, 1, 0 );

    // z axis
    glColor4f( 0, 0, 1, 1 );
    glVertex3f( 0, 0, 0 );
    glVertex3f( 0, 0, 1 );

  glEnd();
  glPopMatrix();
}


/* ===  FUNCTION  ==============================================================
 *         Name:  draw_yelline
 *  Description:  draws a verticle yellow line. used for debugging.
 * ===========================================================================*/
void draw_yelline()
{
  glPushMatrix();
    
  glBegin( GL_LINES );
  
    // y axis
    glColor4f( 1, 1, 0, 1 );
    glVertex3f( 0, 0, 0 );
    glVertex3f( 0, 1, 0 );

  glEnd();
  glPopMatrix();
}

/* ===  FUNCTION  ==============================================================
 *         Name:  my_view_rotate
 *  Description:  rotates the model (so it looks like the camera's rotating
 * ===========================================================================*/
void my_view_rotate(){
  if (g_view_rotate.x!=0){
    g_view_rotate.x = fmod(g_view_rotate.x,360);
    glRotatef(g_view_rotate.x,0,1,0);
  }
}

// =========
// = Audio =
// =========
// Audio callback
int audio_callback( void * outputBuffer, void * inputBuffer, 
      unsigned int bufferSize, double streamTime,
      RtAudioStreamStatus status, void * userData )
{
  SAMPLE * out = (SAMPLE *)outputBuffer;
  SAMPLE * in = (SAMPLE *)inputBuffer;
    
  g_mutex.lock();
    
  // zero out the fft buffer (required only if ZPF > 1)
  memset( g_fftBuff, 0, ZPF * bufferSize * sizeof(float) );
  
  for(size_t i = 0; i < bufferSize; ++i)
  {
    // "static" sine wave for testing
//    g_fftBuff[i] = ::sin( 2 * M_PI * 20.0f * i / MY_SRATE )/2;
//    g_fftBuff[i] += ::sin( 2 * M_PI * 2000.0f * i / MY_SRATE )/2;
    g_fftBuff[i] = in[i];
    out[i] = 0;
  }
  
  // apply window to the buffer of audio
  apply_window( g_fftBuff, g_window, g_buffSize );    

  // compute the fft
  rfft( g_fftBuff, ZPF * bufferSize / 2, FFT_FORWARD );
  
  g_ripple.update_buffs();
  g_mutex.unlock();
  
  return 0;
}


// ========
// = Main =
// ========
// Entry point
int main (int argc, char *argv[])
{
  if (argc>1){
    cerr << "\nUsage: fft\n\tarrow keys control the view\n\tf - toggle in and out of full screen\n\tw - increase number of ripples\n\ts - reset the number of ripples\n\tx - decrease the number of ripples\n\tt - increase the ripples' gain\n\tg - reset the ripples' gain\n\tb - decrease the ripples' gain\n\tm - use mel-scaled frequencies. (experimental)\n\tl - go back to linear scaled frequencies after your mind is blown\n" << endl;
    exit(0);
  }
  // RtAudio config + init
  MyAudio audio(NUM_CHANS, MY_SRATE, &g_buffSize);
  audio.setup(&audio_callback);
  audio.start();

  // allocate the buffer for the fft
  g_fftBuff = new float[g_buffSize * ZPF];
  if ( g_fftBuff == NULL ) {
    cerr << "Something went wrong when creating the fft buffers" << endl;
    exit (1);
  }

  g_ripple=Ripple((complex *) g_fftBuff, g_buffSize, (unsigned int)100, (GLfloat) 0.985, &g_mutex);
  
  // allocate the buffer for the time domain window
  g_window = new float[g_buffSize];
  if ( g_window == NULL ) {
    cerr << "Something went wrong when creating the window" << endl;
    exit (1);
  }

  // create a hanning window
  make_window( g_window, g_buffSize );

  // ============
  // = GL stuff =
  // ============

  // initialize GLUT
  glutInit( &argc, argv );
  // double buffer, use rgb color, enable depth buffer
  glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
  // initialize the window size
  glutInitWindowSize( g_width, g_height );
  // set the window postion
  glutInitWindowPosition( 100, 100 );
  // create the window
  glutCreateWindow( "ripple" );
  //glutEnterGameMode();

  // set the idle function - called when idle
  glutIdleFunc( idleFunc );
  // set the display function - called when redrawing
  glutDisplayFunc( displayFunc );
  // set the reshape function - called when client area changes
  glutReshapeFunc( reshapeFunc );
  // set the keyboard function - called on keyboard events
  glutKeyboardFunc( keyboardFunc );
  // set the mouse function - called on mouse stuff
  glutMouseFunc( mouseFunc );
  // set the special function - called on special keys events (fn, arrows, pgDown, etc)
  glutSpecialFunc( specialFunc );

  // do our own initialization
  initialize();

  // let GLUT handle the current thread from here
  glutMainLoop();

  // if we get here, kill rtaudio
  audio.stop();
  
  return 0;
}
